@@ -53,6 +53,11 @@
 /* autoconfiguration */
 #include "config.h"
 
+/* use l2_ut_[v]sprintf() as a portable [v]snprintf() replacement for debugging */
+#ifdef FSL_DEBUGLOGCODE
+#include "l2_ut_format.h"
+#endif
+
 /* version */
 #define _FSL_VERSION_C_AS_HEADER_
 #include "fsl_version.c"
@@ -61,6 +66,9 @@
 /* pcre static vector size */
 #define OVECSIZE 30
 
+/* maximum atomic write size, maximum log message size */
+#define MAXWRITESIZE 512
+
 /* backward compatibility */
 #ifndef LOG_PRIMASK
 #define LOG_PRIMASK (LOG_EMERG|LOG_ALERT|LOG_CRIT|LOG_ERR|\
@@ -177,48 +185,142 @@ typedef struct {
 
 /* internal global context structure */
 static struct {
-    l2_env_t     *l2_fslenv;
-    l2_channel_t *l2_fslnch;
     l2_env_t     *l2_env;
     l2_channel_t *l2_nch;
+    char         *cpISF;
     levelmap_t   *levelmap;
     int           maskpri;
     int           logopt;
     int           delayopen;
     int           triedopenlog;
-    int           triedopenfsldebug;
-    int           openfsldebugpermanenterror;
-} ctx = { 
-    NULL,
+#ifdef FSL_DEBUGLOGCODE
+    l2_env_t     *fsldebug_l2_env;
+    l2_channel_t *fsldebug_l2_nch;
+    int           fsldebug_transientproblem;
+    int           fsldebug_permanentproblem;
+#endif /* ifdef FSL_DEBUGLOGCODE */
+    char         *fsl_config;
+} ctx = {
     NULL,
     NULL,
     NULL,
     NULL,
     LOG_UPTO(LOG_DEBUG),
     0,
+    TRUE,
+    FALSE,
+#ifdef FSL_DEBUGLOGCODE
+    NULL,
+    NULL,
     FALSE,
     FALSE,
-    FALSE,
-    FALSE
+#endif /* ifdef FSL_DEBUGLOGCODE */
+    "@(#)OSSP fsl config "
+    "cfgdir=" FSL_CFGDIR " "
+    "prefix=" FSL_PREFIX
+    "\n"
+    "@(#)OSSP fsl debug "
+#ifdef FSL_DEBUGLOGCODE
+    "logfile=" FSL_DEBUGLOGFILE " "
+    "logmask=" FSL_DEBUGLOGMASK
+#else /* ifdef FSL_DEBUGLOGCODE */
+    "code omitted"
+#endif /* ifdef FSL_DEBUGLOGCODE */
+    "\n"
 };
 
-/* internal debugging function */
+#ifdef FSL_DEBUGLOGCODE
+static void vfsldebug(l2_level_t level, const char *message, va_list ap)
+{
+    /* sizing cp temporary buffer is max out of
+       strlen "(panic,critical,error,warning,notice,info,trace,debug)" + terminating NUL and
+       strlen "bbb dd HH:MM:SS " + terminating NUL
+       = 56
+     */
+    static char cp[56];
+    static char cpo[MAXWRITESIZE];
+    int cpn;
+    int n;
+    struct stat sb;
+    unsigned int levelmask;
+    time_t t;
+    struct tm *tm;
+    int fd = 0;
+
+    /* determine level mask */
+    if ((n = readlink(FSL_DEBUGLOGMASK, cp, sizeof(cp))) == -1)
+        VCU;
+    if (cp[0] == '(' && cp[n-1] == ')') { /* level mask in round brackets like l2spec */
+        if(l2_util_s2l(&cp[1], n-2, ',', &levelmask) != L2_OK)
+            VCU;
+    }
+    else if (strchr(cp, ',') != NULL) { /* level mask automatic because comma detected */
+        if(l2_util_s2l(&cp[0], n  , ',', &levelmask) != L2_OK)
+            VCU;
+    }
+    else { /* upto level like l2spec */
+        if(l2_util_s2l(&cp[0], n  , 0  , &levelmask) != L2_OK)
+            VCU;
+        levelmask = L2_LEVEL_UPTO(levelmask);
+    }
+
+    if (level & levelmask) {
+
+        cpn = 0;
+
+        /* date and time */
+        t = time(NULL);
+        tm = localtime(&t);
+        if ((n = strftime(cp, sizeof(cp), "%b %d %H:%M:%S ", tm)) == 0)
+            VCU;
+        l2_util_sprintf( &cpo[cpn], MAXWRITESIZE-cpn, "%s", cp);
+        cpn = strlen(cpo);
+
+        /* level */
+        l2_util_sprintf( &cpo[cpn], MAXWRITESIZE-cpn, "<%s> ", l2_util_l2s(cp, sizeof(cp), '\0', level) == L2_OK ? cp : "unknown");
+        cpn = strlen(cpo);
+
+        /* program "name" and pid */
+        l2_util_sprintf( &cpo[cpn], MAXWRITESIZE-cpn, "%s[%lu^%lu] ", ctx.cpISF != NULL ? ctx.cpISF : "unknown", (unsigned long)getpid(), (unsigned long)getppid());
+        cpn = strlen(cpo);
+
+        /* message */
+        l2_util_vsprintf(&cpo[cpn], MAXWRITESIZE-cpn, message, ap);
+        cpn = strlen(cpo);
+        l2_util_sprintf( &cpo[cpn], MAXWRITESIZE-cpn, "\n");
+        cpn = strlen(cpo);
+
+        /* write the log */
+        if (cpn >= 1) {
+            if ((fd = open(FSL_DEBUGLOGFILE, O_WRONLY|O_CREAT|O_APPEND|O_NONBLOCK, 0644)) == -1)
+                VCU;
+            if (fstat(fd, &sb) == -1)
+                VCU;
+            if (sb.st_size >= FSL_DEBUGLOGSTOP)
+                VCU;
+            write(fd, cpo, cpn);
+        }
+    }
+CUS:
+    if (fd)
+        (void)close(fd);
+}
+
 static void fsldebug(l2_level_t level, const char *message, ...)
 {
     va_list ap;
-    char cp[8];
 
     va_start(ap, message);
-    /* note that a L2_ERR_USE is only returned by l2_channel_vlog() if the channel is not yet open */
-    if (ctx.l2_fslnch == NULL || (l2_channel_vlog(ctx.l2_fslnch, level, message, ap) == L2_ERR_USE))
-        if (level & L2_LEVEL_UPTO(L2_LEVEL_ERROR)) {
-            fprintf( stderr, "%s: ", l2_util_l2s(cp, sizeof(cp), '\0', level) == L2_OK ? cp[0]&=0xDF, cp : "Log");
-            vfprintf(stderr, message, ap);
-            fprintf( stderr, "\n");
-    }
+    vfsldebug(level, message, ap);
     va_end(ap);
     return;
 }
+#else /* ifdef FSL_DEBUGLOGCODE */
+static void fsldebug(l2_level_t level, const char *message, ...)
+{
+    return;
+}
+#endif /* ifdef FSL_DEBUGLOGCODE */
 
 /* append contents of a file to buffer */
 static fsl_rc_t appendfiletobuffer(buf_t *buffer, const char *filename)
@@ -283,10 +385,8 @@ static fsl_rc_t readallfiles(buf_t *buffer)
     if (buffer == NULL)
         CU(FSL_ERR_ARG);
 
-    if ((cfgdir = getenv("FSL_CFGDIR")) == NULL)
-        cfgdir = FSL_CFGDIR;
-    if ((prefix = getenv("FSL_PREFIX")) == NULL)
-        prefix = FSL_PREFIX;
+    cfgdir = FSL_CFGDIR;
+    prefix = FSL_PREFIX;
 
     fsldebug(L2_LEVEL_TRACE, "readallfiles() globbing \"%s/%s*\"", cfgdir, prefix);
 
@@ -404,7 +504,7 @@ static int substcapture(const char *inbuf, int inlen, const char **capary, int n
     return n;
 }
 
-/*  process OSSP cfg node tree directives 
+/*  process OSSP cfg node tree directives
     mode=0 processes map/ident,
     mode=1 processes default
  */
@@ -585,7 +685,7 @@ static fsl_rc_t processcfg(cfg_t *cfg, char *cpISF, int mode)
                 }
                 if (mapto == -1) {
                     fsldebug(L2_LEVEL_ERROR, "processcfg: trying to map to unknown l2 level \"%s\"", cfgargtoka[2]); CU(FSL_ERR_CUS); }
-                
+
                 ctx.levelmap[mapfrom].l2 = l2level2string[mapto].level;
                 fsldebug(L2_LEVEL_DEBUG, "processcfg: map levelmap[%10s/%d].l2 = l2level2string[%10s/%d].level = 0x%.8lx", cfgargtoka[1], mapfrom, cfgargtoka[2], mapto, (unsigned long)l2level2string[mapto].level);
             }
@@ -613,68 +713,6 @@ CUS:
     return rc;
 }
 
-static void closefsllog()
-{
-    if (ctx.l2_fslnch != NULL) {
-        l2_channel_destroy(ctx.l2_fslnch);
-        ctx.l2_fslnch = NULL;
-    }
-    if (ctx.l2_fslenv != NULL) {
-        l2_env_destroy(ctx.l2_fslenv);
-        ctx.l2_fslenv = NULL;
-    }
-    ctx.triedopenfsldebug = FALSE;
-}
-
-static void openfsldebug()
-{
-    int rc = FSL_OK;
-    char *argl2spec;
-    l2_result_t   l2rv;
-    char *cp; /* scratch variable */
-    l2_channel_t *ch; /* scratch variable */
-
-    if (ctx.triedopenfsldebug || ctx.openfsldebugpermanenterror)
-        return;
-
-    /* properly prepare for repeated execution */
-    ctx.triedopenfsldebug = TRUE;
-
-    /* create OSSP l2 environment for fsl itself (internal logging) */
-    if ((argl2spec = getenv("FSL_DEBUG")) == NULL)
-        argl2spec = FSL_DEBUG;
-    if (strlen(argl2spec) > 0) {
-        if ((l2rv = l2_env_create(&ctx.l2_fslenv)) != L2_OK) {
-            cp = l2_env_strerror(ctx.l2_fslenv, l2rv); fsldebug(L2_LEVEL_ERROR, "openlog: failed to create L2 environment (%d) for fsl", l2rv); CU(FSL_ERR_INT); }
-        if ((l2rv = l2_env_levels(ctx.l2_fslenv, L2_LEVEL_ALL, L2_LEVEL_NONE)) != L2_OK) {
-            cp = l2_env_strerror(ctx.l2_fslenv, l2rv); fsldebug(L2_LEVEL_ERROR, "openlog: logging failed to set global logging level defaults %s(%d) for fsl", cp, l2rv); CU(FSL_ERR_INT); }
-        if ((l2rv = l2_env_formatter(ctx.l2_fslenv, 'D', l2_util_fmt_dump, NULL)) != L2_OK) {
-            cp = l2_env_strerror(ctx.l2_fslenv, l2rv); fsldebug(L2_LEVEL_ERROR, "openlog: logging failed to register dump formatter %s(%d) for fsl", cp, l2rv); CU(FSL_ERR_INT); }
-        if ((l2rv = l2_env_formatter(ctx.l2_fslenv, 'S', l2_util_fmt_string, NULL)) != L2_OK) {
-            cp = l2_env_strerror(ctx.l2_fslenv, l2rv); fsldebug(L2_LEVEL_ERROR, "openlog: logging failed to register string formatter %s(%d) for fsl", cp, l2rv); CU(FSL_ERR_INT); }
-        if ((l2rv = l2_env_formatter(ctx.l2_fslenv, 'm', l2_util_fmt_errno, NULL)) != L2_OK) {
-            cp = l2_env_strerror(ctx.l2_fslenv, l2rv); fsldebug(L2_LEVEL_ERROR, "openlog: logging failed to register errno formatter %s(%d) for fsl", cp, l2rv); CU(FSL_ERR_INT); }
-        if ((l2rv = l2_channel_create(&ctx.l2_fslnch, ctx.l2_fslenv, "noop")) != L2_OK) {
-            cp = l2_env_strerror(ctx.l2_fslenv, l2rv); fsldebug(L2_LEVEL_ERROR, "openlog: logging failed to create noop channel; %s(%d) for fsl", cp, l2rv); CU(FSL_ERR_INT); }
-        if ((l2rv = l2_spec(&ch, ctx.l2_fslenv, "%s", argl2spec)) != L2_OK) {
-            cp = l2_env_strerror(ctx.l2_fslenv, l2rv); fsldebug(L2_LEVEL_ERROR, "openlog: logging failed to create stream from spec %s(%d) for fsl", cp, l2rv);
-            if (l2rv == L2_ERR_ARG)
-                CU(FSL_ERR_ARG);
-            else
-                CU(FSL_ERR_INT);
-        }
-        if ((l2rv = l2_channel_link(ctx.l2_fslnch, L2_LINK_CHILD, ch, NULL)) != L2_OK) {
-            cp = l2_env_strerror(ctx.l2_fslenv, l2rv); fsldebug(L2_LEVEL_ERROR, "openlog: logging failed to link child channel %s(%d) for fsl", cp, l2rv); CU(FSL_ERR_INT); }
-        if ((l2rv = l2_channel_open(ctx.l2_fslnch)) != L2_OK) {
-            cp = l2_env_strerror(ctx.l2_fslenv, l2rv); fsldebug(L2_LEVEL_ERROR, "openlog: logging failed to open channel stream %s(%d) for fsl", cp, l2rv); CU(FSL_ERR_INT); }
-    }
-CUS:
-    if (rc != FSL_OK)
-        closefsllog();
-    if (rc == FSL_ERR_ARG)
-        ctx.openfsldebugpermanenterror = TRUE;
-}
-
 /* POSIX API function openlog(3) */
 void openlog(const char *ident, int logopt, int facility)
 {
@@ -687,7 +725,6 @@ void openlog(const char *ident, int logopt, int facility)
     int   i;  /* scratch variable */
     char *cpIdent;
     char *cpFacility;
-    char *cpISF;
     l2_result_t   l2rv;
 
     /* initialization */
@@ -696,27 +733,29 @@ void openlog(const char *ident, int logopt, int facility)
     buf.used = 0;
     buf.size = 0;
     cpIdent = NULL;
-    cpISF = NULL;
 
     /* properly prepare for repeated execution */
     closelog();
     ctx.triedopenlog = TRUE;
 
-    /* create OSSP l2 environment for fsl itself (internal logging) */
-    openfsldebug();
+    /* remember logopt */
+    ctx.logopt = logopt;
+
+    /* handle delayed open logopt value */
+    if (ctx.logopt & LOG_NDELAY)
+        ctx.delayopen = FALSE;
 
     /* tracing */
-    fsldebug(L2_LEVEL_TRACE, "openlog() ident=\"%s\", logopt=0x%.8lx, facility=0x%.8lx; caught by %s", ident, logopt, facility, fsl_version.v_gnu);
+    fsldebug(L2_LEVEL_TRACE, "fsl in openlog(3) ident=%s%s%s, logopt=0x%.8lx, facility=0x%.8lx; caught by %s", ident == NULL ? "" : "\"", ident == NULL ? "NULL" : ident, ident == NULL ? "" : "\"" , logopt, facility, fsl_version.v_gnu);
 
-    /* remember logopt and handle unsupported values */
-    ctx.logopt = logopt;
-    if (logopt & LOG_CONS)
+    /* handle unsupported logopt values */
+    if (ctx.logopt & LOG_CONS)
         fsldebug(L2_LEVEL_WARNING, "openlog: ignore unsupported LOG_CONS");
 #ifdef LOG_PERROR
-    if (logopt & LOG_PERROR)
+    if (ctx.logopt & LOG_PERROR)
         fsldebug(L2_LEVEL_WARNING, "openlog: ignore unsupported LOG_PERROR (use OSSP l2 channel \"fd(filedescriptor=2)\" to emulate)");
 #endif
-    if (logopt & LOG_PID)
+    if (ctx.logopt & LOG_PID)
         fsldebug(L2_LEVEL_WARNING, "openlog: ignore unsupported LOG_PID (use OSSP l2 formatter %%P in prefix channel to emulate)");
 
     /* create default sysloglevel to l2_level mapping */
@@ -728,7 +767,7 @@ void openlog(const char *ident, int logopt, int facility)
     for (i = 0; sysloglevel2string[i].string != NULL; i++) {
         ctx.levelmap[i].syslog = sysloglevel2string[i].level;
         ctx.levelmap[i].l2     = sysloglevel2string[i].deflevelmap;
-        fsldebug(L2_LEVEL_DEBUG, "openlog: ctx.levelmap[%d].syslog = 0x%.8lx, ctx.levelmap[%d].l2 = 0x%.8lx", 
+        fsldebug(L2_LEVEL_DEBUG, "openlog: ctx.levelmap[%d].syslog = 0x%.8lx, ctx.levelmap[%d].l2 = 0x%.8lx",
                  i, (unsigned long)ctx.levelmap[i].syslog, i, (unsigned long)ctx.levelmap[i].l2);
     }
 
@@ -756,12 +795,12 @@ void openlog(const char *ident, int logopt, int facility)
             break;
         }
     }
-    if ((cpISF = (char *)malloc(strlen(cpIdent) + 1 + strlen(cpFacility) + 1)) == NULL) {
+    if ((ctx.cpISF = (char *)malloc(strlen(cpIdent) + 1 + strlen(cpFacility) + 1)) == NULL) {
         fsldebug(L2_LEVEL_ERROR, "openlog: malloc() failed"); CU(1); }
-    cpISF[0] = '\0';
-    strcat(cpISF, cpIdent);
-    strcat(cpISF, "/");
-    strcat(cpISF, cpFacility);
+    ctx.cpISF[0] = '\0';
+    strcat(ctx.cpISF, cpIdent);
+    strcat(ctx.cpISF, "/");
+    strcat(ctx.cpISF, cpFacility);
 
     /* read configuration file(s) into buffer */
     if ((rv = readallfiles(&buf)) != FSL_OK) {
@@ -774,16 +813,15 @@ void openlog(const char *ident, int logopt, int facility)
         (void)cfg_error(cfg, cfgrv, &cp); fsldebug(L2_LEVEL_ERROR, "openlog: cfg_import() failed with error %s (%d)", cp, cfgrv); CU(1); }
 
     /*  process OSSP cfg node tree "map" and "ident" directives */
-    if ((rv = processcfg(cfg, cpISF, 0)) != FSL_OK && (rv != FSL_NOIDENT)) {
+    if ((rv = processcfg(cfg, ctx.cpISF, 0)) != FSL_OK && (rv != FSL_NOIDENT)) {
         fsldebug(L2_LEVEL_ERROR, "openlog: processcfg() failed with an unrecoverable error (%d)", rv); CU(1); }
 
     /*  optionally process OSSP cfg node tree "default" directives */
-    if ((rv == FSL_NOIDENT) && ((rv = processcfg(cfg, cpISF, 1)) != FSL_OK)) {
+    if ((rv == FSL_NOIDENT) && ((rv = processcfg(cfg, ctx.cpISF, 1)) != FSL_OK)) {
         fsldebug(L2_LEVEL_ERROR, "openlog: processcfg() failed with an unrecoverable error (%d)", rv); CU(1); }
 
     /*  open logging now or prepare for delayed open */
-    if (logopt & LOG_NDELAY) {
-        ctx.delayopen = TRUE;
+    if (~logopt & LOG_NDELAY) {
         fsldebug(L2_LEVEL_TRACE, "openlog: logopt LOG_NDELAY delays open of L2 channel tree until first message is being logged");
     }
     else {
@@ -794,8 +832,10 @@ void openlog(const char *ident, int logopt, int facility)
 
     CU(0);
 CUS:
-    if (cpISF != NULL)
-        free(cpISF);
+    if (ctx.cpISF != NULL) {
+        free(ctx.cpISF);
+        ctx.cpISF = NULL;
+    }
     if (cpIdent != NULL)
         free(cpIdent);
     if (cfg != NULL)
@@ -811,8 +851,9 @@ CUS:
 /* faked POSIX API function closelog(3) */
 void closelog(void)
 {
-    fsldebug(L2_LEVEL_TRACE, "closelog()");
-    closefsllog();
+    /* tracing */
+    fsldebug(L2_LEVEL_TRACE, "fsl in closelog(3)");
+
     if (ctx.l2_nch != NULL) {
         l2_channel_destroy(ctx.l2_nch);
         ctx.l2_nch = NULL;
@@ -825,10 +866,15 @@ void closelog(void)
         free(ctx.levelmap);
         ctx.levelmap = NULL;
     }
+    if (ctx.cpISF != NULL) {
+        free(ctx.cpISF);
+        ctx.cpISF = NULL;
+    }
     ctx.maskpri = LOG_UPTO(LOG_DEBUG);
     ctx.logopt = 0;
-    ctx.delayopen = FALSE;
+    ctx.delayopen = TRUE;
     ctx.triedopenlog = FALSE;
+
     return;
 }
 
@@ -837,7 +883,9 @@ int setlogmask(int maskpri)
 {
     int oldmask;
 
-    fsldebug(L2_LEVEL_TRACE, "setlogmask() maskpri=0x%.8lx", maskpri);
+    /* tracing */
+    fsldebug(L2_LEVEL_TRACE, "fsl in setlogmask(3) maskpri=0x%.8lx", maskpri);
+
     oldmask = ctx.maskpri;
     if (maskpri != 0)
         ctx.maskpri = maskpri;
@@ -849,6 +897,9 @@ void syslog(int priority, const char *message, ...)
 {
     va_list args;
 
+    /* tracing */
+    fsldebug(L2_LEVEL_TRACE, "fsl in syslog(3); go ahead using vsyslog(3)");
+
     /* wrap around vsyslog(3) */
     va_start(args, message);
     vsyslog(priority, message, args);
@@ -867,8 +918,11 @@ void vsyslog(int priority, const char *fmt, va_list args)
     int i;
     l2_result_t l2rv;
     char *cp;
-    
-    /* check for previous omitted attempt of initialization */
+
+    /* tracing */
+    fsldebug(L2_LEVEL_TRACE, "fsl in vsyslog(3) fmt=%s%s%s ...", fmt == NULL ? "" : "\"", fmt == NULL ? "NULL" : fmt, fmt == NULL ? "" : "\"");
+
+    /* check for omitted openlog(3) */
     if (ctx.l2_nch == NULL && !ctx.triedopenlog)
         openlog("fsl", 0, LOG_SYSLOG);
 
@@ -877,7 +931,7 @@ void vsyslog(int priority, const char *fmt, va_list args)
         return;
 
     /* check for delayed open */
-    if ((ctx.logopt & LOG_NDELAY) && (ctx.delayopen == TRUE)) {
+    if ((~ctx.logopt & LOG_NDELAY) && (ctx.delayopen == TRUE)) {
         if ((l2rv = l2_channel_open(ctx.l2_nch)) != L2_OK) {
             cp = l2_env_strerror(ctx.l2_env, l2rv); fsldebug(L2_LEVEL_ERROR, "vsyslog: logging failed to open channel stream %s(%d) delayed", cp, l2rv);
             closelog();
