@@ -1,10 +1,11 @@
 /*
-**  L2 - OSSP Logging Library
-**  Copyright (c) 2001-2002 The OSSP Project (http://www.ossp.org/)
-**  Copyright (c) 2001-2002 Cable & Wireless Deutschland (http://www.cw.com/de/)
+**  OSSP l2 - Flexible Logging
+**  Copyright (c) 2001-2003 Cable & Wireless Deutschland GmbH
+**  Copyright (c) 2001-2003 The OSSP Project (http://www.ossp.org/)
+**  Copyright (c) 2001-2003 Ralf S. Engelschall <rse@engelschall.com>
 **
-**  This file is part of OSSP L2, a flexible logging library which
-**  can be found at http://www.ossp.org/pkg/l2/.
+**  This file is part of OSSP l2, a flexible logging library which
+**  can be found at http://www.ossp.org/pkg/lib/l2/.
 **
 **  Permission to use, copy, modify, and distribute this software for
 **  any purpose with or without fee is hereby granted, provided that
@@ -30,7 +31,7 @@
 
 #include "l2_p.h"
 
-/* 
+/*
  * A channel is the central object for a logging stream. It is
  * implemented by a framework (the code implemented here) which provides
  * the channel API and which is equal for all channels and a particular
@@ -309,8 +310,8 @@ l2_result_t l2_channel_open(l2_channel_t *ch)
         rv = ch->handler.open(&ch->context, ch);
     else
         rv = L2_OK_PASS;
-   
-    /* optionally pass operation downstream */ 
+
+    /* optionally pass operation downstream */
     if (rv == L2_OK_PASS) {
         rv = L2_OK;
         chD = NULL;
@@ -367,8 +368,8 @@ l2_result_t l2_channel_write(l2_channel_t *ch, l2_level_t level, const char *buf
         rv = ch->handler.write(&ch->context, ch, level, buf, bufsize);
     else
         rv = L2_OK_PASS;
-   
-    /* optionally pass operation downstream */ 
+
+    /* optionally pass operation downstream */
     if (rv == L2_OK_PASS) {
         rv = L2_OK;
         chD = NULL;
@@ -400,8 +401,8 @@ l2_result_t l2_channel_flush(l2_channel_t *ch)
         rv = ch->handler.flush(&ch->context, ch);
     else
         rv = L2_OK_PASS;
-   
-    /* optionally pass operation downstream */ 
+
+    /* optionally pass operation downstream */
     if (rv == L2_OK_PASS) {
         rv = L2_OK;
         chD = NULL;
@@ -423,7 +424,7 @@ l2_result_t l2_channel_close(l2_channel_t *ch)
     /* argument sanity check */
     if (ch == NULL)
         return L2_ERR_ARG;
-    
+
     /* make sure channel is in state "opened" */
     if (ch->state != L2_CHSTATE_OPENED)
         return L2_ERR_USE;
@@ -433,8 +434,8 @@ l2_result_t l2_channel_close(l2_channel_t *ch)
         rv = ch->handler.close(&ch->context, ch);
     else
         rv = L2_OK_PASS;
-   
-    /* optionally pass operation downstream */ 
+
+    /* optionally pass operation downstream */
     if (rv == L2_OK_PASS) {
         rv = L2_OK;
         chD = NULL;
@@ -454,8 +455,10 @@ l2_result_t l2_channel_close(l2_channel_t *ch)
 l2_result_t l2_channel_destroy(l2_channel_t *ch)
 {
     l2_result_t rv;
-    l2_result_t rvD;
+    l2_result_t rvD;   /* downstream */
     l2_channel_t *chD;
+    l2_result_t rvL;   /* lookahead */
+    l2_channel_t *chL;
 
     /* argument sanity check */
     if (ch == NULL)
@@ -471,14 +474,21 @@ l2_result_t l2_channel_destroy(l2_channel_t *ch)
         rv = ch->handler.destroy(&ch->context, ch);
     else
         rv = L2_OK_PASS;
-   
-    /* optionally pass operation downstream */ 
+
+    /* optionally pass operation downstream */
     if (rv == L2_OK_PASS) {
         rv = L2_OK;
         chD = NULL;
-        while (l2_channel_downstream(ch, &chD) == L2_OK)
-            if ((rvD = l2_channel_destroy(chD)) != L2_OK)
-                rv = rvD;
+        if (l2_channel_downstream(ch, &chD) == L2_OK) {
+            chL = chD;
+            do {
+                rvL = l2_channel_downstream(ch, &chL);
+                if ((rvD = l2_channel_destroy(chD)) != L2_OK)
+                    rv = rvD;
+                if (rvL == L2_OK)
+                    chD = chL;
+            } while ((rv == L2_OK) && (rvL == L2_OK));
+        }
     }
 
     /* free channel structure */
@@ -528,7 +538,7 @@ static void l2_channel_vlog_format(
     /* iterate over all configured L2 formatters */
     for (i = 0; i < L2_MAX_FORMATTERS && env->formatters[i].cb != NULL; i++) {
         if (env->formatters[i].id == cId) {
-            rv = env->formatters[i].cb(env->formatters[i].ctx, cId, cpParam, 
+            rv = env->formatters[i].cb(env->formatters[i].ctx, cId, cpParam,
                                        cpBuf, nBufLenMax, npOutLen, apArgs);
             vfmt->data[1].i = (int)rv;
             if (rv == L2_OK) {
@@ -550,7 +560,7 @@ l2_result_t l2_channel_vlog(l2_channel_t *ch, l2_level_t level, const char *fmt,
     l2_env_t *env;
 
     /* argument sanity check */
-    if (ch == NULL || level == 0 || fmt == NULL || ap == NULL)
+    if (ch == NULL || level == 0 || fmt == NULL)
         return L2_ERR_ARG;
 
     /* make sure only a single level is specified */
